# Copyright 2024 The AI Edge Model Explorer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import json
import os
from dataclasses import asdict, dataclass, field
from typing import Union

from .utils import remove_none

Num = Union[float, int]
NodeData = Union['ModelNodeData', 'GraphNodeData']


# The node data for a model.
#
# It stores a dict that indexes `GraphNodeData` by graph ids.
@dataclass
class ModelNodeData:
  """The node data for a model."""

  # Graph node data indexed by graph ids.
  graphsData: dict[str, 'GraphNodeData']

  def save_to_file(self, path: str, indent: Union[int, None] = 2) -> None:
    """Writes the data into the given file."""
    abs_path = os.path.abspath(os.path.expanduser(path))
    with open(abs_path, 'w') as f:
      data = self.to_json_string(indent=indent)
      f.write(data)

  def to_json_string(self, indent: Union[int, None] = None) -> str:
    data = {k: remove_none(asdict(v))
            for (k, v) in self.graphsData.items()}
    return json.dumps(data, indent=indent)


@dataclass
class GraphNodeData:
  """Node data for a single graph."""

  # Node data values indexed by node keys.
  #
  # The key could be:
  # - Any of the output tensor names of a node. It will be used to match a
  #   node's `tensor_name` key in its `outputsMetadata`. See graph_builder.py
  #   for more info.
  # - The node id specified in `GraphNode`. See graph_builder.py for more info.
  results: dict[str, 'NodeDataResult']

  # Thresholds that define various ranges and the corresponding node styles
  # (e.g. node bg color) to be applied for that range.
  #
  # Take the following thresholds as an example:
  #
  # [
  #   {value: 10, bgColor: 'red'}
  #   {value: 50, bgColor: 'blue'}
  #   {value: 100, bgColor: 'yellow'}
  # ]
  #
  # This means:
  # - Node data with value <=10 have "red" background color.
  # - Node data with value >10 and <=50 have "blue" background color.
  # - Node data with value >50 and <=100 have "yellow" background color.
  # - Node data with value >100 have no background color (white).
  #
  # (optional)
  thresholds: list['ThresholdItem'] = field(default_factory=list)

  # A gradient that defines the stops (from 0 to 1) and the associated colors.
  # A stop value 0 corresponds to the minimum value in `results`, and a stop
  # value 1 corresponds to the maximum value in results. Stops for 0 and 1
  # should always be provided.
  #
  # When color-coding a node, the system uses the node's data value to
  # calculate a corresponding stop and interpolates between gradient colors.
  #
  # This field takes precedence over the `thresholds` field above.
  #
  # (optional)
  gradient: list['GradientItem'] = field(default_factory=list)

  def save_to_file(self, path: str, indent: Union[int, None] = 2) -> None:
    """Writes the data into the given file."""
    abs_path = os.path.abspath(os.path.expanduser(path))
    with open(abs_path, 'w') as f:
      data = self.to_json_string(indent=indent)
      f.write(data)

  def to_json_string(self, indent: Union[int, None] = None) -> str:
    data = remove_none(asdict(self))
    return json.dumps(data, indent=indent)


@dataclass
class NodeDataResult:
  """A single result data corresponding to a node."""

  # A numeric value.
  value: Num

  # The background color to render the corresponding node with.
  #
  # This could be any CSS color. It takes precedence over the color calculated
  # from `GraphNodeData.thresholds` and `GraphNodeData.gradient`.
  #
  # (optional)
  bgColor: Union[str, None] = None

  # The text label color to render the corresponding node with.
  #
  # This could be any CSS color. It takes precedence over the color calculated
  # from `GraphNodeData.thresholds` and `GraphNodeData.gradient`.
  #
  # (optional)
  textColor: Union[str, None] = None


@dataclass
class ThresholdItem:
  """A threshold item with the upperbound value and its corresponding colors."""

  # The value of the threshold.
  value: Num

  # The background color to render for node whose value falls into the segment
  # defined by this threshold item.
  #
  # This could be any CSS color.
  bgColor: str

  # The text label color to render for node whose value falls into the segment
  # defined by this threshold item.
  #
  # This could be any CSS color.
  #
  # (optional)
  textColor: Union[str, None] = None


@dataclass
class GradientItem:
  """A gradient item with the stop and its corresponding colors."""

  # A number from 0 to 1.
  #
  # 0 and 1 corresponds to the minimum/maximum value of the all the values in
  # data.
  stop: Num

  # The background color to render at this stop.
  #
  # Only support hex color (e.g. #aabb00) or color names (e.g. red).
  #
  # (optional)
  bgColor: Union[str, None] = None

  # The text color to render at this stop.
  #
  # Only support hex color (e.g. #aabb00) or color names (e.g. red).
  #
  # (optional)
  textColor: Union[str, None] = None



# Copyright 2024 The AI Edge Model Explorer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import argparse

from . import server
from .config import ModelExplorerConfig
from .consts import DEFAULT_HOST, DEFAULT_PORT

parser = argparse.ArgumentParser(
    prog='model-explorer',
    description='A modern model graph visualizer and debugger',
    formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument('model_paths',
                    nargs='?',
                    help='comma separated model file paths')
parser.add_argument('--host',
                    default=DEFAULT_HOST,
                    help='host of the server')
parser.add_argument('--port',
                    default=DEFAULT_PORT,
                    type=int,
                    help='port of the server')
parser.add_argument('--node_data_paths',
                    help='comma separated node data file paths')
parser.add_argument('--no_open_in_browser',
                    action='store_true',
                    help='Don\'t open the web app in browser after server starts')
parser.add_argument('--extensions',
                    help='comma separated extension module names')
parser.add_argument('--cors_host',
                    help='the host of the Access-Control-Allow-Origin header')
args = parser.parse_args()


def main():
  """Entry point for the command line version of model explorer."""

  model_paths: list[str] = []
  if args.model_paths is not None and args.model_paths != '':
    model_paths = [x.strip() for x in args.model_paths.split(',')]

  node_data_paths: list[str] = []
  if args.node_data_paths is not None and args.node_data_paths != '':
    node_data_paths = [x.strip() for x in args.node_data_paths.split(',')]

  extensions: list[str] = []
  if args.extensions is not None:
    extensions = [x.strip() for x in args.extensions.split(',')]

  # Construct config.
  config = ModelExplorerConfig()
  for model_path in model_paths:
    config.add_model_from_path(model_path)
  for node_data_path in node_data_paths:
    config.add_node_data_from_path(node_data_path)

  server.start(host=args.host,
               port=args.port,
               config=config,
               extensions=extensions,
               cors_host=args.cors_host,
               no_open_in_browser=args.no_open_in_browser)



# Copyright 2024 The AI Edge Model Explorer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import os
import shutil
from dataclasses import asdict
from importlib import import_module
from typing import Any, Dict, Union

from .adapter_runner import AdapterRunner
from .consts import MODULE_NAME
from .extension_class_processor import ExtensionClassProcessor
from .registered_extension import RegisteredExtension
from .singleton import Singleton
from .types import AdapterCommand
from .utils import convert_adapter_response


class ExtensionManager(object, metaclass=Singleton):
  BUILTIN_ADAPTER_MODULES: list[str] = [
      '.builtin_tflite_flatbuffer_adapter',
      '.builtin_tflite_mlir_adapter',
      '.builtin_tf_mlir_adapter',
      '.builtin_tf_direct_adapter',
      '.builtin_graphdef_adapter',
      '.builtin_pytorch_exportedprogram_adapter',
      '.builtin_mlir_adapter',
  ]

  CACHED_REGISTERED_EXTENSIONS: Dict[str, RegisteredExtension] = {}

  def __init__(self, custom_extension_modules: list[str] = []):
    # For custom extensions (i.e. non-built-in extensions), we load their "main"
    # module by default.
    self.custom_extension_modules = [
        f'{x}.main' for x in custom_extension_modules]
    self.extensions: list[RegisteredExtension] = []
    self.adapter_runner: AdapterRunner = AdapterRunner()

  def load_extensions(self) -> None:
    """Loads all extensions."""
    self.extensions = []
    self._import_extensions()

  def get_extensions_metadata(self) -> list:
    """Get metadata for all extensions."""
    exts = [{**asdict(ext.metadata), 'type': ext.type}
            for ext in self.extensions]
    exts.append({
        'fileExts': ['json'],
        'type': 'adapter',
        'id': 'builtin_json',
        'name': 'JSON adapter',
        'description': 'Convert graphs json data file or tfjs model.',
    })
    return exts

  def run_cmd(self, cmd: Any) -> Any:
    """Runs the given command."""
    # Get the extension.
    extension_id = cmd['extensionId']
    extension = self._get_extension_by_id(extension_id)
    if extension is None:
      raise Exception(f'Extension "{extension_id}" not found')

    # Run.
    #
    # Adapter.
    if extension.type == 'adapter':
      resp = self.adapter_runner.run_adapter(
          extension=extension, cmd=cmd)
      return convert_adapter_response(resp)

    return {}

  def cleanup(self, cmd: Any):
    # Get the extension.
    extension_id = cmd['extensionId']
    extension = self._get_extension_by_id(extension_id)

    if extension is not None:
      # Adapter
      if extension.type == 'adapter':
        # Delete the file if it is marked "deleteAfterConversion".
        adapter_cmd: AdapterCommand = cmd
        if adapter_cmd['deleteAfterConversion']:
          model_path = adapter_cmd['modelPath']
          model_dir = os.path.dirname(model_path)
          shutil.rmtree(model_dir, ignore_errors=True)

  def _import_extensions(self):
    # Built-in pywrapped c++ extensions + custom extensions.
    for module in ExtensionManager.BUILTIN_ADAPTER_MODULES + self.custom_extension_modules:
      module_full_name = f'{MODULE_NAME}{module}'

      # Get the registered extension from cache if it has already been
      # registered.
      if module_full_name in ExtensionManager.CACHED_REGISTERED_EXTENSIONS:
        self.extensions.append(
            ExtensionManager.CACHED_REGISTERED_EXTENSIONS[module_full_name])
      # Import the extension module if it has not been registered.
      else:
        try:
          import_module(module, MODULE_NAME)
        except Exception as err:
          print(f'! Failed to load extension module "{module}":')
          print(err)
          print()
          continue

        if ExtensionClassProcessor.extension_class is not None:
          extension = RegisteredExtension(metadata=ExtensionClassProcessor.extension_class.metadata,
                                          type=ExtensionClassProcessor.extension_type,
                                          ext_class=ExtensionClassProcessor.extension_class)
          self.extensions.append(extension)
          ExtensionManager.CACHED_REGISTERED_EXTENSIONS[module_full_name] = extension

  def _get_extension_by_id(self, id: str) -> Union[RegisteredExtension, None]:
    matches = [ext for ext in self.extensions if ext.metadata.id == id]
    if len(matches) > 0:
      return matches[0]
    return None



# Copyright 2024 The AI Edge Model Explorer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from importlib.metadata import version

from . import graph_builder, node_data_builder
from .adapter import Adapter, AdapterMetadata
from .apis import config, visualize, visualize_from_config, visualize_pytorch
from .consts import PACKAGE_NAME
from .types import ModelExplorerGraphs

# Default 'exports'.
__all__ = ['config', 'visualize', 'visualize_pytorch', 'visualize_from_config',
           'Adapter', 'AdapterMetadata', 'ModelExplorerGraphs', 'graph_builder',
           'node_data_builder']

__version__ = version(PACKAGE_NAME)



# Copyright 2024 The AI Edge Model Explorer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from dataclasses import dataclass, field
from typing import Union


@dataclass
class Graph:
  """A Model Explorer graph."""

  # The id of the graph.
  id: str

  # A list of nodes in the graph.
  nodes: list['GraphNode'] = field(default_factory=list)


@dataclass
class GraphNode:
  """A node in a graph."""
  # The unique id of the node. It only needs to be unique within the graph.
  id: str

  # The label of the node, displayed on the node in the model graph.
  label: str

  # Define the node's namespace hierarchy in the form of a "path" (e.g., a/b/c).
  # Don't include the node label as the last component of the namespace.
  # This enables the visualizer to create nested layers for easier navigation.
  #
  # For example, for three nodes with the following label and namespace data:
  # - N1: a/b
  # - N2: a/b
  # - N3: a
  #
  # The visualizer will first show a collapsed layer labeled 'a'. After the
  # layer is expanded (by user clicking on it), it will reveal node N3, and
  # another collapsed layer labeled 'b'. After the layer 'b' is expanded, it
  # will reveal two nodes N1 and N2 inside.
  #
  # (optional)
  namespace: str = ''

  # Ids of subgraphs that this node goes into.
  #
  # The ids should point to other Graph objects returned by your adapter
  # extension.  Once set, users will be able to click this node, pick a subgraph
  # from a drop-down list, and see the visualization for the selected subgraph.
  #
  # (optional)
  subgraphIds: list[str] = field(default_factory=list)

  # The attributes of the node.
  #
  # They will be displayd in the side panel when a node is selected:
  #
  # - attr1_key
  #   attr1_value
  #
  # - attr2_key
  #   attr2_value
  # ...
  #
  # (optional)
  attrs: list['KeyValue'] = field(default_factory=list)

  # A list of incoming edges.
  incomingEdges: list['IncomingEdge'] = field(default_factory=list)

  # Metadata for outputs.
  #
  # Typically an adapter would store metadata for output tensors here.
  #
  # They will be displayed in the side panel when a node is selected:
  #
  # Output 0:
  #   - attr1_key: attr1_value
  #   - attr2_key: attr2_value
  #
  # Output 1:
  #   - attr1_key: attr1_value
  #   - attr2_key: attr2_value
  # ...
  #
  # (optional)
  outputsMetadata: list['MetadataItem'] = field(default_factory=list)

  # Metadata for inputs.
  #
  # The following data is stored in this field:
  #
  # - __tensor_tag: identify the name of a node's input tensors.
  # - quantization: the quantization parameter for tflite models when using
  #   tflite direct adapter.
  #
  # (optional)
  inputsMetadata: list['MetadataItem'] = field(default_factory=list)

  # The style of the node.
  #
  # This will overwrite the default node styles.
  #
  # (optional)
  style: Union['GraphNodeStyle', None] = None


@dataclass
class IncomingEdge:
  """An incoming edge of a graph node."""
  # The id of the source node.
  sourceNodeId: str

  # The id of the output from the source node that this edge goes out of.
  #
  # (optional)
  sourceNodeOutputId: str = '0'

  # The id of the input from the target node that this edge connects to.
  #
  # (optional)
  targetNodeInputId: str = '0'


@dataclass
class KeyValue:
  """A key-value pair"""
  key: str
  value: str


@dataclass
class MetadataItem:
  """A single output metadata item for a specific output id."""
  # The id of the input/output this metadata item belongs to.
  id: str

  # The attributes for the input/output.
  #
  # (optional)
  attrs: list[KeyValue] = field(default_factory=list)


@dataclass
class GraphNodeStyle:
  """Style of the node."""
  # The background color of the node.
  #
  # (optional)
  backgroundColor: str = ''

  # The border color of the node.
  #
  # (optional)
  borderColor: str = ''

  # The border color of the node when hovered.
  #
  # (optional)
  hoveredBorderColor: str = ''


@dataclass
class GraphCollection:
  # The label of the collection.
  #
  # It will be appended to the file name to distinguish different graph
  # collections in the graph selector.
  label: str

  # A list of graphs in this collection.
  graphs: list[Graph] = field(default_factory=list)



# Copyright 2024 The AI Edge Model Explorer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import json
import os
from typing import Any, Callable, Tuple, TypedDict, Union
from urllib.parse import quote

import torch
from typing_extensions import NotRequired

from .node_data_builder import GraphNodeData, ModelNodeData, NodeData
from .pytorch_exported_program_adater_impl import \
    PytorchExportedProgramAdapterImpl
from .types import ModelExplorerGraphs

ModelSource = TypedDict(
    'ModelSource', {'url': str, 'adapterId': NotRequired[str]})

EncodedUrlData = TypedDict(
    'EncodedUrlData', {
        'models': list[ModelSource],
        'nodeData': NotRequired[list[str]]})


class ModelExplorerConfig:
  """Stores the data to be visualized in Model Explorer."""

  def __init__(self) -> None:
    self.model_sources: list[ModelSource] = []
    self.graphs_list: list[ModelExplorerGraphs] = []
    self.node_data_sources: list[str] = []
    self.node_data_list: list[NodeData] = []

  def add_model_from_path(self, path: str, adapterId: str = '') -> 'ModelExplorerConfig':
    """Adds a model path to the config.

    Args:
      path: the model path to add.
      adapterId: the id of the adapter to use. Default is empty string meaning
          it will use the default adapter.
    """
    # Get the absolute path (after expanding home dir path "~").
    abs_model_path = os.path.abspath(os.path.expanduser(path))

    # Construct model source and add it.
    model_source: ModelSource = {'url': abs_model_path}
    if adapterId != '':
      model_source['adapterId'] = adapterId
    self.model_sources.append(model_source)

    return self

  def add_model_from_pytorch(self,
                             name: str,
                             exported_program: torch.export.ExportedProgram) -> 'ModelExplorerConfig':
    """Adds the given pytorch model.

    Args:
      name: the name of the model for display purpose.
      exported_program: the ExportedProgram from torch.export.export.
    """
    # Convert the given model to model explorer graphs.
    print('Converting pytorch model to model explorer graphs...')
    adapter = PytorchExportedProgramAdapterImpl(exported_program)
    graphs = adapter.convert()
    graphs_index = len(self.graphs_list)
    self.graphs_list.append(graphs)

    # Construct model source.
    #
    # The model source has a special format, in the form of:
    # graphs://{name}/{graphs_index}
    model_source: ModelSource = {'url': f'graphs://{name}/{graphs_index}'}
    self.model_sources.append(model_source)

    return self

  def add_node_data_from_path(self, path: str) -> 'ModelExplorerConfig':
    """Adds node data file to the config.

    Args:
      path: the path of the node data json file to add.
    """
    # Get the absolute path (after expanding home dir path "~").
    abs_model_path = os.path.abspath(os.path.expanduser(path))

    self.node_data_sources.append(abs_model_path)

    return self

  def add_node_data(self, name: str, node_data: NodeData) -> 'ModelExplorerConfig':
    """Adds the given node data object.

    Args:
      name: the name of the NodeData for display purpose.
      node_data: the NodeData object to add.
    """
    node_data_index = len(self.node_data_list)
    self.node_data_list.append(node_data)

    # Construct sources.
    #
    # The node data source has a special format, in the form of:
    # node_data://{name}//{index}
    self.node_data_sources.append(f'node_data://{name}/{node_data_index}')
    return self

  def to_url_param_value(self) -> str:
    """Converts the config to the url param value."""
    # Construct url data.
    encoded_url_data: EncodedUrlData = {
        'models': self.model_sources
    }

    if self.node_data_sources:
      encoded_url_data['nodeData'] = self.node_data_sources

    # Return its json string.
    return quote(json.dumps(encoded_url_data))

  def get_model_explorer_graphs(self, index: int) -> ModelExplorerGraphs:
    return self.graphs_list[index]

  def get_node_data(self, index: int) -> NodeData:
    return self.node_data_list[index]

  def has_data_to_encode_in_url(self) -> bool:
    return len(self.model_sources) > 0 or len(self.node_data_sources) > 0



# Copyright 2024 The AI Edge Model Explorer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from typing import Union

import torch

from . import server
from .config import ModelExplorerConfig
from .consts import DEFAULT_COLAB_HEIGHT, DEFAULT_HOST, DEFAULT_PORT


def config() -> ModelExplorerConfig:
  """Create a new config object."""
  return ModelExplorerConfig()


def visualize(
        model_paths: Union[str, list[str]] = [],
        host=DEFAULT_HOST,
        port=DEFAULT_PORT,
        colab_height=DEFAULT_COLAB_HEIGHT) -> None:
  """Starts the ME local server and visualizes the models by the given paths.

  Args:
    model_paths: A model path or a list of model paths to visualize.
    host: The host of the server. Default to localhost.
    port: The port of the server. Default to 8080.
    colab_height: The height of the embedded iFrame when running in colab.
  """
  # Construct config.
  cur_config = config()
  model_paths_list = model_paths
  if isinstance(model_paths, str):
    model_paths_list = [model_paths]
  for model_path in model_paths_list:
    cur_config.add_model_from_path(path=model_path)

  # Start server.
  server.start(
      host=host,
      port=port,
      config=cur_config,
      colab_height=colab_height)


def visualize_pytorch(
        name: str,
        exported_program: torch.export.ExportedProgram,
        host=DEFAULT_HOST,
        port=DEFAULT_PORT,
        colab_height=DEFAULT_COLAB_HEIGHT) -> None:
  """Visualizes a pytorch model.

  Args:
    name: The name of the model for display purpose.
    exported_program: The ExportedProgram from torch.export.export.
    host: The host of the server. Default to localhost.
    port: The port of the server. Default to 8080.
    colab_height: The height of the embedded iFrame when running in colab.
  """
  # Construct config.
  cur_config = config()
  cur_config.add_model_from_pytorch(name, exported_program=exported_program)

  # Start server.
  server.start(
      host=host,
      port=port,
      config=cur_config,
      colab_height=colab_height)


def visualize_from_config(
        config: Union[ModelExplorerConfig, None] = None,
        host=DEFAULT_HOST,
        port=DEFAULT_PORT,
        cors_host: Union[str, None] = None,
        no_open_in_browser: bool = False,
        colab_height=DEFAULT_COLAB_HEIGHT) -> None:
  """Visualizes with a config.

  Args:
    config: the object that stores the models to be visualized.
    host: The host of the server. Default to localhost.
    port: The port of the server. Default to 8080.
    cors_host: The value of the Access-Control-Allow-Origin header. The header
        won't be present if it is None.
    no_open_in_browser: Don't open the web app in browser after server starts.
    colab_height: The height of the embedded iFrame when running in colab.
  """
  # Start server.
  server.start(
      host=host,
      port=port,
      config=config,
      cors_host=cors_host,
      no_open_in_browser=no_open_in_browser,
      colab_height=colab_height)
